"""
Run Annalist server.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os, os.path
import sys
import importlib
import subprocess
import signal
import time
import shutil

import logging
log = logging.getLogger(__name__)

from utils.SetcwdContext            import ChangeCurrentDir
from utils.SuppressLoggingContext   import SuppressLogging
from annalist                       import __version__

from .                              import am_errors
from .am_settings                   import (
    am_get_settings, am_get_site_settings, am_get_site
    )
from .am_getargvalue                import getarg, getargvalue

def am_runserver(annroot, userhome, options):
    """
    Run Annalist production server asynchronously; writes procdss id to stdout.

    This uses the gunicorn HTTP/WSGI server.
    Provide HTTPS access by proxying via Apache or Nginx.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    sitedirectory = sitesettings.BASE_SITE_DIR
    pidfilename   = os.path.join(sitedirectory, "annalist.pid")
    try:
        with open(pidfilename, "r") as pidfile:
            pid = int(pidfile.readline())
            print("Server already started with pid %d"%(pid,), file=sys.stderr)
            if options.force:
                print("Stopping pid %d"%(pid,), file=sys.stderr)
                os.kill(pid, signal.SIGTERM)
            else:
                print("Use '--force' to force restart", file=sys.stderr)
                return am_errors.AM_SERVERALREADYRUN
    except IOError as e:
        # No saved process id - continue
        pass
    except OSError as e:
        # Process Id not found for kill
        print("Process pid %d not found"%(pid,), file=sys.stderr)
        pass
    status = am_errors.AM_SUCCESS
    with ChangeCurrentDir(annroot):
        gunicorn_command = (
            "gunicorn --workers=1 --threads=2 "+
            "    --bind=0.0.0.0:8000 "+
            "    --env DJANGO_SETTINGS_MODULE=%s "%(settings.modulename,)+
            "    --env ANNALIST_KEY=%s "%(sitesettings.SECRET_KEY,)+
            "    --access-logfile %s "%(sitesettings.ACCESS_LOG_PATH,)+
            "    --error-logfile  %s "%(sitesettings.ERROR_LOG_PATH,)+
            "    --timeout 300 "+
            "    annalist_site.wsgi:application"+
            "")
        log.debug("am_runserver subprocess: %s"%gunicorn_command)
        p   = subprocess.Popen(gunicorn_command.split())
        pid = p.pid
        with open(pidfilename, "w") as pidfile:
            pidfile.write(str(pid)+"\n")
        time.sleep(1.0) # Allow server to start and log initial messages
        print(str(pid), file=sys.stdout)
        log.debug("am_runserver subprocess pid: %s"%pid)
    return status

def am_stopserver(annroot, userhome, options):
    """
    Stop Annalist production server.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    sitedirectory = sitesettings.BASE_SITE_DIR
    pidfilename   = os.path.join(sitedirectory, "annalist.pid")
    status = am_errors.AM_SUCCESS
    try:
        with open(pidfilename, "r") as pidfile:
            pid = int(pidfile.readline())
            print("Stopping pid %d"%(pid,), file=sys.stderr)
            os.kill(pid, signal.SIGTERM)
        os.remove(pidfilename)
    except IOError as e:
        # print("PID file %s not found (%s)"%(pidfilename, e), file=sys.stderr)
        print("No server running", file=sys.stderr)
        return am_errors.AM_NOSERVERPIDFILE
    except OSError as e:
        print("Process %d not found (%s)"%(pid, e), file=sys.stderr)
        return am_errors.AM_PIDNOTFOUND
    return status

def am_pidserver(annroot, userhome, options):
    """
    Display running Annalist server PID on stdout

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if PID is displayed, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    sitedirectory = sitesettings.BASE_SITE_DIR
    pidfilename   = os.path.join(sitedirectory, "annalist.pid")
    status = am_errors.AM_SUCCESS
    try:
        with open(pidfilename, "r") as pidfile:
            pid = int(pidfile.readline())
            print("%d"%(pid,), file=sys.stdout)
    except IOError as e:
        # print("PID file %s not found (%s)"%(pidfilename, e), file=sys.stderr)
        print("No server running", file=sys.stderr)
        return am_errors.AM_NOSERVERPIDFILE
    return status

def am_rundevserver(annroot, userhome, options):
    """
    Run Annalist developent server.

    This uses the Django developm,ent server (via manage.py).
    For production deployment, use `runserver`.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with ChangeCurrentDir(annroot):
        cmd = "runserver 0.0.0.0:8000"
        subprocess_command = "django-admin %s --pythonpath=%s --settings=%s"%(cmd, annroot, settings.modulename)
        log.debug("am_rundevserver subprocess: %s"%subprocess_command)
        status = subprocess.call(subprocess_command.split())
        log.debug("am_rundevserver subprocess status: %s"%status)
    return status

def am_serverlog(annroot, userhome, options):
    """
    Print name of Annalist server log to standard output.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    logfilename = sitesettings.ANNALIST_LOG_PATH
    print(logfilename)
    return status

def am_accesslog(annroot, userhome, options):
    """
    Print name of WSGI access log to standard output.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    logfilename = sitesettings.ACCESS_LOG_PATH
    print(logfilename)
    return status

def am_errorlog(annroot, userhome, options):
    """
    Print name of WSGI access log to standard output.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    logfilename = sitesettings.ERROR_LOG_PATH
    print(logfilename)
    return status

def am_sitedirectory(annroot, userhome, options):
    """
    Print name of Annalist site directory to standard output

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    sitedirectory = sitesettings.BASE_SITE_DIR
    print(sitedirectory)
    # with open(sitedirectory, "r") as logfile:
    #     shutil.copyfileobj(logfile, sys.stdout)
    return status

def am_settingsmodule(annroot, userhome, options):
    """
    Print name of Annalist settings module to standard output

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    settingsmodule = settings.modulename
    print(settingsmodule)
    return status

def am_settingsfile(annroot, userhome, options):
    """
    Print name of Annalist settings file to standard output

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    settingsfile, ext = os.path.splitext(sitesettings.__file__)
    print(settingsfile)
    return status

def am_settingsdir(annroot, userhome, options):
    """
    Print name of Annalist settings file to standard output

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    settingsdir, file = os.path.split(sitesettings.__file__)
    print(settingsdir)
    return status

def am_version(annroot, userhome, options):
    """
    Print software version string to standard output.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    print(sitesettings.ANNALIST_VERSION)
    # with open(logfilename, "r") as logfile:
    #     shutil.copyfileobj(logfile, sys.stdout)
    return status

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for collection data viewing and editing.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import re
import json
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                            import settings
from django.db                              import models
from django.http                            import QueryDict
from django.core.urlresolvers               import resolve, reverse
from django.contrib.auth.models             import User
from django.test                            import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client                     import Client
from django.template                        import Template, Context

from utils.SuppressLoggingContext           import SuppressLogging

import annalist
from annalist                               import layout
from annalist.identifiers                   import RDF, RDFS, ANNAL
from annalist.util                          import valid_id

from annalist.models.site                   import Site
from annalist.models.sitedata               import SiteData
from annalist.models.collection             import Collection
from annalist.models.annalistuser           import AnnalistUser

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site, init_annalist_test_coll, resetSitedata
    )
from .entity_testutils import (
    site_dir, collection_dir,
    site_view_url, collection_view_url, collection_edit_url, 
    collection_entity_view_url,
    collection_create_values,
    create_user_permissions,
    create_test_user,
    context_view_field
    )
from .entity_testentitydata import (
    entity_url, entitydata_edit_url, entitydata_list_type_url,
    default_fields, default_label, default_comment, error_label,
    layout_classes
    )
from .entity_testcolldata import (
    collectiondata_url, collectiondata_resource_url,
    collectiondata_view_url, collectiondata_view_resource_url,
    collectiondata_value_keys, collectiondata_load_keys,
    collectiondata_create_values, collectiondata_values, collectiondata_read_values,
    coll_view_form_data
    )
from .entity_testuserdata import (
    user_view_form_data
    )


#   -----------------------------------------------------------------------------
#
#   Test collection data access and editing interfaces
#
#   -----------------------------------------------------------------------------

class CollectionDataEditViewTest(AnnalistTestCase):
    """
    Tests for collection data edit view
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        # Login and permissions
        create_test_user(
            self.testsite.site_data_collection(),
            # self.testcoll, 
            "testuser", "testpassword",
            user_permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG", "ADMIN"]
            )
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        resetSitedata(scope="all")
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _check_collection_data_values(self, coll_id=None, coll_label=None, coll_descr=None):
        """
        Helper function checks content of annalist collection data
        """
        self.assertTrue(Collection.exists(self.testsite, coll_id))
        c = Collection.load(self.testsite, coll_id)
        self.assertEqual(c.get_id(), coll_id)
        self.assertEqual(c.get_view_url_path(), collection_view_url(coll_id="testcoll"))
        v = collectiondata_values(
            coll_id=coll_id, coll_label=coll_label, coll_descr=coll_descr
            )
        self.assertDictionaryMatch(c.get_values(), v)
        return c

    def _check_annalist_user_perms(self, user_id, user_perms):
        self.assertTrue(AnnalistUser.exists(self.testcoll, user_id))
        u = AnnalistUser.load(self.testcoll, user_id)
        self.assertEqual(u.get_id(), user_id)
        self.assertEqual(u[ANNAL.CURIE.user_permission], user_perms)
        return

    #   -----------------------------------------------------------------------------
    #   Form rendering and access tests
    #   -----------------------------------------------------------------------------

    def test_get_collection_data_form_rendering(self):
        u = entitydata_edit_url("new", layout.SITEDATA_ID, layout.COLL_TYPEID, view_id="Collection_view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        field_vals = default_fields(
            coll_id=layout.SITEDATA_ID, 
            type_id=layout.COLL_TYPEID, 
            entity_id="00000001", 
            sotware_ver=annalist.__version_data__,
            tooltip1a=context_view_field(r.context, 0, 0)['field_tooltip'],
            tooltip1b=context_view_field(r.context, 0, 1)['field_tooltip'],
            tooltip2 =context_view_field(r.context, 1, 0)['field_tooltip'],
            tooltip3 =context_view_field(r.context, 2, 0)['field_tooltip'],
            tooltip4 =context_view_field(r.context, 3, 0)['field_tooltip'],
            tooltip5a=context_view_field(r.context, 4, 0)['field_tooltip'],
            tooltip5b=context_view_field(r.context, 4, 1)['field_tooltip'],
            tooltip6a=context_view_field(r.context, 5, 0)['field_tooltip'],
            tooltip6b=context_view_field(r.context, 5, 1)['field_tooltip'],
            tooltip7 =context_view_field(r.context, 6, 0)['field_tooltip'],
            )
        formrow1a = """
            <div class="small-12 medium-6 columns" title="%(tooltip1a)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Id</span>
                    </div>
                    <div class="%(input_classes)s">
                        <input type="text" size="64" name="entity_id" 
                                   placeholder="(entity id)" 
                                   value="00000001" />
                    </div>
                </div>
            </div>
            """%field_vals(width=6)
        formrow1b = """
            <div class="small-12 medium-6 columns" title="%(tooltip1b)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>S/W version</span>
                    </div>
                    <div class="%(input_classes)s">
                        <span>&nbsp;</span>
                    </div>
                </div>
            </div>
            """%field_vals(width=6)            
        formrow2 = """
            <div class="small-12 columns" title="%(tooltip2)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Label</span>
                    </div>
                    <div class="%(input_classes)s">
                        <input type="text" size="64" name="Entity_label" 
                               placeholder="(label)"
                               value="" />
                    </div>
                </div>
            </div>
            """%field_vals(width=12)
        formrow3 = """
            <div class="small-12 columns" title="%(tooltip3)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Comment</span>
                    </div>
                    <div class="%(input_classes)s">
                        <textarea cols="64" rows="6" name="Entity_comment" 
                                  class="small-rows-4 medium-rows-8"
                                  placeholder="(description)"
                                  >
                        </textarea>
                    </div>
                </div>
            </div>
            """%field_vals(width=12)
        formrow4 = """
            <div class="small-12 columns" title="%(tooltip4)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Parent</span>
                    </div>
                    <div class="%(input_classes)s">
                        <select name="Coll_parent">
                            <option value="" selected="selected">(site)</option>
                            <option value="_coll/_annalist_site">Annalist data notebook test site</option>
                            <option value="_coll/coll1">Collection coll1</option>
                            <option value="_coll/coll2">Collection coll2</option>
                            <option value="_coll/coll3">Collection coll3</option>
                            <option value="_coll/testcoll">Collection testcoll</option>
                        </select>
                    </div>
                </div>
            </div>
            """%field_vals(width=12)
        formrow5a = """
             <div class="small-12 medium-6 columns" title="%(tooltip5a)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Default list</span>
                    </div>
                    <div class="%(input_classes)s">
                        <span>&nbsp;</span>
                    </div>
                </div>
            </div>
           """%field_vals(width=6)
        formrow5b = """
            <div class="small-12 medium-6 columns" title="%(tooltip5b)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Default view</span>
                    </div>
                    <div class="%(input_classes)s">
                        <span>&nbsp;</span>
                    </div>
                </div>
            </div>
            """%field_vals(width=6)            
        formrow6a = """
             <div class="small-12 medium-6 columns" title="%(tooltip6a)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Default view type</span>
                    </div>
                    <div class="%(input_classes)s">
                        <span>&nbsp;</span>
                    </div>
                </div>
            </div>
           """%field_vals(width=6)
        formrow6b = """
            <div class="small-12 medium-6 columns" title="%(tooltip6b)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Default view entity</span>
                    </div>
                    <div class="%(input_classes)s">
                        <span>&nbsp;</span>
                    </div>
                </div>
            </div>
            """%field_vals(width=6)            
        formrow7 = """
            <div class="small-12 columns" title="%(tooltip7)s">
                <div class="row view-value-row">
                    <div class="%(label_classes)s">
                        <span>Collection metadata</span>
                    </div>
                    <div class="%(input_classes)s">
                        <textarea cols="64" rows="6" name="Coll_comment" 
                                  class="small-rows-4 medium-rows-8" 
                                  placeholder="(annal:meta_comment)"
                                  >
                        </textarea>
                    </div>
                </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content)
        self.assertContains(r, formrow1a, html=True)
        self.assertContains(r, formrow1b, html=True)
        self.assertContains(r, formrow2,  html=True)
        self.assertContains(r, formrow3,  html=True)
        self.assertContains(r, formrow4,  html=True)
        self.assertContains(r, formrow5a, html=True)
        self.assertContains(r, formrow5b, html=True)
        self.assertContains(r, formrow6a, html=True)
        self.assertContains(r, formrow6b, html=True)
        # self.assertContains(r, formrow_viewdata, html=True)
        return

    # Collection data view
    def test_get_collection_data_view(self):
        collection_url = collection_view_url(coll_id="testcoll")
        u = collectiondata_url(coll_id="testcoll")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        "testcoll")
        self.assertEqual(r.context['orig_id'],          "testcoll")
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        self.assertEqual(
            r.context['entity_data_ref'],      
            collection_url+layout.COLL_META_REF
            )
        self.assertEqual(
            r.context['entity_data_ref_json'], 
            collection_url+layout.COLL_META_REF+"?type=application/json"
            )
        return

    # Collection data content negotiation
    def test_get_collection_data_json(self):
        """
        Request collection data as JSON-LD
        """
        collection_url = collection_view_url(coll_id="testcoll")
        u = collectiondata_url(coll_id="testcoll")
        r = self.client.get(u, HTTP_ACCEPT="application/ld+json")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        v = r['Location']
        self.assertEqual(v, TestHostUri+collection_url+layout.COLL_META_REF)
        r = self.client.get(v)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return

    # Collection data view metadata access
    def test_get_collection_data_view_metadata(self):
        u = collectiondata_resource_url(coll_id="testcoll")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        colldata = json.loads(r.content)
        expected = (
            { "@id":                    "../"
            , "@type":                  [ "annal:Collection" ]
            , "@context":               [ {"@base": layout.META_COLL_BASE_REF}, "coll_context.jsonld" ]
            , "annal:id":               "testcoll"
            , "annal:type_id":          layout.COLL_TYPEID
            , "annal:type":             "annal:Collection"
            , "rdfs:label":             "Collection testcoll"
            , "rdfs:comment":           "Description of Collection testcoll"
            , "annal:software_version": annalist.__version_data__
            })
        self.assertEqual(colldata, expected)
        return

    # collection named view context
    def test_get_named_view(self):
        u = collectiondata_view_url(coll_id="testcoll", action="view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        "testcoll")
        self.assertEqual(r.context['orig_id'],          "testcoll")
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        return

    # collection named view edit
    def test_get_named_edit(self):
        u = collectiondata_view_url(coll_id="testcoll", action="edit")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        "testcoll")
        self.assertEqual(r.context['orig_id'],          "testcoll")
        self.assertEqual(r.context['action'],           "edit")
        self.assertEqual(r.context['continuation_url'], "")
        return

    # collection named view edit post update
    def test_post_named_edit(self):
        u = collectiondata_view_url(coll_id="testcoll", action="edit")
        f = coll_view_form_data(coll_id="testcoll", action="edit")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   TestBasePath+"/c/_annalist_site/d/_coll/")
        # Check updated collection data
        self._check_collection_data_values(coll_id="testcoll")
        return

    # collection named view metadata access
    def test_get_named_view_metadata(self):
        u = collectiondata_view_resource_url(coll_id="testcoll")
        log.debug("test_get_named_view_metadata: collectiondata_view_resource_url: %s"%(u,))
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        colldata = json.loads(r.content)
        expected = (
            { "@id":                    "../"
            , "@type":                  [ "annal:Collection" ]
            , "@context":               [ {"@base": layout.META_COLL_BASE_REF}, "coll_context.jsonld" ]
            , "annal:id":               "testcoll"
            , "annal:type_id":          layout.COLL_TYPEID
            , "annal:type":             "annal:Collection"
            , "rdfs:label":             "Collection testcoll"
            , "rdfs:comment":           "Description of Collection testcoll"
            , "annal:software_version": annalist.__version_data__
            })
        self.assertEqual(colldata, expected)
        return

    #   -----------------------------------------------------------------------------
    #   Form response tests
    #   -----------------------------------------------------------------------------

    def test_edit_collection_metadata(self):
        # This test performs a GET to retrieve values used in a form,
        # then a POST to save updated collection metadata.
        # This test is intended to test a problem encountered with updated 
        # entity copying logic that needs to take special account of collection
        # entities being presented as offspring of the site while being stored
        # as part of a collection.
        #
        coll_id = "testcoll"
        self.assertTrue(Collection.exists(self.testsite, coll_id))
        c = Collection.load(self.testsite, coll_id)
        self.assertEqual(c.get_id(), coll_id)
        self.assertEqual(c.get_view_url_path(), collection_view_url(coll_id="testcoll"))
        # GET collection metadata form data, and test values
        u = collectiondata_url(coll_id="testcoll")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        "testcoll")
        self.assertEqual(r.context['orig_id'],          "testcoll")
        self.assertEqual(r.context['orig_coll'],        layout.SITEDATA_ID)
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        orig_coll = r.context['orig_coll']
        # Assemble and POST form data to =updated collection metadata
        new_label = "Updated collection metadata"
        f = coll_view_form_data(
            coll_id="testcoll", 
            action="edit", 
            coll_label=new_label,
            # orig_coll="None"
            orig_coll=layout.SITEDATA_ID
            )
        u = collectiondata_view_url(coll_id="testcoll", action="edit")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   TestBasePath+"/c/_annalist_site/d/_coll/")
        # Check updated collection data
        self._check_collection_data_values(coll_id="testcoll", coll_label=new_label)
        return

    def test_post_copy_user(self):
        # The main purpose of this test is to check that user permissions are 
        # saved properly
        self.assertFalse(AnnalistUser.exists(self.testcoll, "copyuser"))
        f = user_view_form_data(
            action="copy", orig_id="_default_user_perms",
            user_id="copyuser",
            user_name="User copyuser",
            user_uri="mailto:copyuser@example.org",
            user_permissions="VIEW CREATE UPDATE DELETE",
            orig_coll=layout.SITEDATA_ID
            )
        u = entitydata_edit_url(
            "copy", "testcoll", "_user", entity_id="_default_user_perms", view_id="User_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'], TestBasePath+"/c/testcoll/d/_user/")
        # Check that new record type exists
        self.assertTrue(AnnalistUser.exists(self.testcoll, "copyuser"))
        self._check_annalist_user_perms("copyuser", ["VIEW", "CREATE", "UPDATE", "DELETE"])
        return

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Renderer and value mapper for URI value displayed as a link.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from utils.py3porting                           import is_string, to_unicode

from annalist.views.fields.render_base          import RenderBase
from annalist.views.fields.render_fieldvalue    import (
    RenderFieldValue,
    get_field_edit_value, get_field_view_value
    )

from django.template    import Template, Context

#   ----------------------------------------------------------------------------
#
#   Link URI value mapping
#
#   ----------------------------------------------------------------------------


class URILinkValueMapper(RenderBase):
    """
    Value mapper class for URI string presented for viewing as a clickable link
    """

    @classmethod
    def encode(cls, data_value):
        """
        Encodes link value as string
        """
        return data_value or ""

    @classmethod
    def decode(cls, field_value):
        """
        Decodes a URI string value as a link
        """
        if is_string(field_value):
            field_value = field_value.strip()
        return field_value or ""


#   ----------------------------------------------------------------------------
#
#   Link URI field renderers
#
#   ----------------------------------------------------------------------------

class uri_link_view_renderer(object):

    def render(self, context):
        """
        Render link for viewing.
        """
        linkval = URILinkValueMapper.encode(get_field_view_value(context, ""))
        # log.info("uri_link_view_renderer: linkval %r (orig)"%(linkval,))
        common_prefixes = (
            [ "http://", "https://"
            , "file:///", "file://localhost/", "file://"
            , "mailto:"]
            )
        textval = linkval
        for p in common_prefixes:
            if is_string(linkval) and linkval.startswith(p):
                textval = linkval[len(p):]
                break
        if ":" in linkval:
            link_pref, link_path = linkval.split(":", 1)
            if "collection" not in context:
                log.warning("uri_link_view_renderer: no key 'collection' in context")
                # log.error("@@@@@@@@@@@@@@@@@@@@@")
                # for k in context.flatten():
                #     hidden_fields = (
                #         [ "fields", "row_bound_fields", "repeat_bound_fields"
                #         , "help_text", "help_markdown"
                #         , "forloop", "f", "block", "view_choices"
                #         , "LANGUAGES"
                #         ])
                #     if k not in hidden_fields:
                #         log.error("    @@@ %s: %r"%(k, context[k]))
            else:
                link_vocab = context["collection"].cache_get_vocab(link_pref)
                if link_vocab:
                    linkval = link_vocab.get_uri() + link_path
        # log.info("uri_link_view_renderer: linkval %r (final)"%(linkval,))
        return '''<a href="%s" target="_blank">%s</a>'''%(linkval, textval)

class uri_link_edit_renderer(object):

    def __init__(self):
        self._template = Template(
            '''<input type="text" size="64" name="{{repeat_prefix}}{{field.description.field_name}}" '''+
                   '''placeholder="{{field.description.field_placeholder}}" '''+
                   '''value="{{field.field_edit_value}}" />'''
        )
        return

    def render(self, context):
        """
        Render link for editing
        """
        return self._template.render(context)

def get_uri_link_renderer():
    """
    Return field renderer object for URI link values
    """
    return RenderFieldValue("uri_link",
        view_renderer=uri_link_view_renderer(), 
        edit_renderer=uri_link_edit_renderer(),
        )

# End.



"""
Root settings module for Annalist.

Default submodule is selected by manage.py or am_runserver.py

To specify an alternative:
  python manage.py rundevserver --settings=annalist_site.settings.personal
  python manage.py runserver --settings=annalist_site.settings.personal
or
  export DJANGO_SETTINGS_MODULE=annalist_site.settings.personal

See also:
  http://www.rdegges.com/the-perfect-django-settings-file/

"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

import sys

import logging
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)
log.debug("Python path:    "+repr(sys.path))

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
This module is used to cache per-collection field information.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from annalist                       import layout
from annalist.exceptions            import Annalist_Error
from annalist.identifiers           import ANNAL, RDFS

from annalist.models.collectionentitycache  import (
    Cache_Error, CollectionEntityCacheObject, CollectionEntityCache
    )
from annalist.models.closurecache           import ClosureCache
from annalist.models.recordfield            import RecordField

#   ---------------------------------------------------------------------------
# 
#   Field-cache object class
# 
#   ---------------------------------------------------------------------------

class CollectionFieldCacheObject(CollectionEntityCacheObject):
    """
    This class is a field definition cache for a specified collection.

    It extends class CollectionEntityCacheObject with field-specific logic; notably
    overriding method _load_entity with additional logic to maintain a superproperty
    closure cache, and methods to access that cache.
    """
    def __init__(self, coll_id, entity_cls=RecordField):
        """
        Initialize a cache object for a specified collection.

        coll_id         Collection id with which the field cache is associated.
        """
        super(CollectionFieldCacheObject, self).__init__(coll_id, entity_cls)
        self._superproperty_closure = ClosureCache(coll_id, ANNAL.CURIE.superproperty_uri)
        return

    def _load_entity(self, coll, field_entity):
        """
        Internal helper method loads field data to cache.
        Also updates superproperty closure cache.

        Returns True if new field was added.
        """
        field_id     = field_entity.get_id()
        property_uri = field_entity.get_property_uri()
        field_parent = field_entity.get_parent().get_id()
        field_data   = field_entity.get_save_values()
        add_field    = super(CollectionFieldCacheObject, self)._load_entity(
                            coll, field_entity, entity_uri=property_uri
                            )
        if add_field:
            # Add relations for superproperty references from the new property URI
            for superproperty_obj in field_data.get(ANNAL.CURIE.superproperty_uri, []):
                superproperty_uri = superproperty_obj["@id"]
                self._superproperty_closure.add_rel(property_uri, superproperty_uri)
            # Also add relations for references *to* the new property URI
            for try_subproperty_obj in self.get_all_entities(coll):
                sub_superp_objs = try_subproperty_obj.get(ANNAL.CURIE.superproperty_uri, [])
                sub_superp_uris = (
                    [ sub_superp_obj["@id"] for sub_superp_obj in sub_superp_objs ]
                    )
                if property_uri in sub_superp_uris:
                    subproperty_uri = try_subproperty_obj.get(ANNAL.CURIE.property_uri, None)
                    if subproperty_uri:
                        self._superproperty_closure.add_rel(subproperty_uri, property_uri)
        return add_field

    def _drop_entity(self, coll, field_id):
        """
        Override method that drops an entity from the cache, to also remove references
        from the superproperty closure cache.

        Returns the field entity removed, or None if not found.
        """
        field_entity = super(CollectionFieldCacheObject, self)._drop_entity(coll, field_id)
        if field_entity:
            property_uri = field_entity.get_property_uri()
            self._superproperty_closure.remove_val(property_uri)
        return field_entity

    def get_superproperty_uris(self, property_uri):
        """
        Returns all superproperty URIs for a specified property URI.

        Returns all superproperty URIs, even those for which there 
        is no defined field entity.
        """
        return self._superproperty_closure.fwd_closure(property_uri)

    def get_subproperty_uris(self, property_uri):
        """
        Returns all subproperty URIs for a specified property URI.

        Returns all subproperty URIs, even those for which there 
        is no defined field entity.
        """
        return self._superproperty_closure.rev_closure(property_uri)

    def get_superproperty_fields(self, coll, property_uri):
        """
        Returns all superproperties for a specified property URI.

        This method returns only those superproperties that are defined as entities.
        """
        self._load_entities(coll)
        for st_uri in self.get_superproperty_uris(property_uri):
            st = self.get_entity_from_uri(coll, st_uri)
            if st:
                yield st
        return

    def get_subproperty_fields(self, coll, property_uri):
        """
        Returns all subproperties for a specified property URI.

        This method returns only those subproperties that are defined as entities.
        """
        self._load_entities(coll)
        for st_uri in self.get_subproperty_uris(property_uri):
            st = self.get_entity_from_uri(coll, st_uri)
            if st:
                yield st
        return

    def remove_cache(self):
        """
        Close down and release all collection field cache data
        """
        # log.debug("@@@@remove field cache %r"%(self.get_coll_id(),))
        super(CollectionFieldCacheObject, self).remove_cache()
        self._superproperty_closure.remove_cache()
        self._superproperty_closure = None
        return

#   ---------------------------------------------------------------------------
# 
#   Collection field-cache class
# 
#   ---------------------------------------------------------------------------

class CollectionFieldCache(CollectionEntityCache):
    """
    This class manages field cache objects over multiple collections
    """
    def __init__(self):
        """
        Initialize.

        Initializes a value cache cache with no per-collection data.
        """
        super(CollectionFieldCache, self).__init__(CollectionFieldCacheObject, RecordField)
        return

    # Collection field cache allocation and access methods

    def set_field(self, coll, field_entity):
        """
        Save a new or updated field definition
        """
        return self.set_entity(coll, field_entity)

    def remove_field(self, coll, field_id):
        """
        Remove field from collection field cache.

        Returns the field entity removed if found, or None if not defined.
        """
        return self.remove_entity(coll, field_id)

    def get_field(self, coll, field_id):
        """
        Retrieve a field description for a given field Id.

        Returns a field object for the specified collection and field Id.
        """
        return self.get_entity(coll, field_id)

    def get_field_from_uri(self, coll, field_uri):
        """
        Retrieve a field description for a given property URI.

        Returns a field object for the specified collection and property URI.
        """
        return self.get_entity_from_uri(coll, field_uri)

    def get_all_field_ids(self, coll, altscope=None):
        """
        Returns all fields currently available for a collection in the indicated scope.
        Default scope is fields defined directly in the indicated collection.
        """
        return self.get_all_entity_ids(coll, altscope=altscope)

    def get_all_fields(self, coll, altscope=None):
        """
        Returns all fields currently available for a collection in the indicated scope.
        Default scope is fields defined directly in the indicated collection.
        """
        return self.get_all_entities(coll, altscope=altscope)

    def get_superproperty_fields(self, coll, field_uri):
        """
        Returns all superproperties for a specieid property URI.
        """
        field_cache = self._get_cache(coll)
        return field_cache.get_superproperty_fields(coll, field_uri)

    def get_subproperty_fields(self, coll, field_uri):
        """
        Returns all subproperties for a specieid property URI.
        """
        field_cache = self._get_cache(coll)
        return field_cache.get_subproperty_fields(coll, field_uri)

    def get_superproperty_uris(self, coll, field_uri):
        """
        Returns all superproperties for a specieid property URI.
        """
        field_cache = self._get_cache(coll)
        return field_cache.get_superproperty_uris(field_uri)

    def get_subproperty_uris(self, coll, field_uri):
        """
        Returns all subproperties for a specieid property URI.
        """
        field_cache = self._get_cache(coll)
        return field_cache.get_subproperty_uris(field_uri)

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
This module provides an object cacheing framework for arbitrary Python values.
The intent is thatall cacghe logic can be isolated, and may be re-implemented
using a network cache faclity such as MemCache or Redis.

The present implementation assumes a single-process, multi-threaded environment
and interlocks cache accesses to avoid possible cache-related race conditions.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2019, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import traceback
import threading
import contextlib

from annalist.exceptions            import Annalist_Error

#   ===================================================================
#
#   Error class
#
#   ===================================================================

class Cache_Error(Annalist_Error):
    """
    Class for errors raised by cache methods.
    """
    def __init__(self, value=None, msg="Cache_error (objectcache)"):
        super(Cache_Error, self).__init__(value, msg)
        return

#   ===================================================================
#
#   Cache creation and discovery
#
#   ===================================================================

globalcachelock  = threading.Lock() # Used to interlock creation/deletion of caches
objectcache_dict = {}               # Initial empty set of object caches
objectcache_tb   = {}

def get_cache(cachekey):
    """
    This function locates or creates an object cache.

    cachekey    is a hashable value that uniquely identifies the required cache
                (e.g. a string or URI).

    Returns the requested cache object, which may be created on-the-fly.
    """
    with globalcachelock:
        if cachekey not in objectcache_dict:
            objectcache_dict[cachekey] = ObjectCache(cachekey)
        objectcache = objectcache_dict[cachekey]    # Copy value while lock acquired
    return objectcache

def remove_cache(cachekey):
    """
    This function removes a cache from the set of object  caches

    cachekey    is a hashable value that uniquely identifies the required cache
                (e.g. a string or URI).
    """
    # log.debug("objectcache.remove_cache %r"%(cachekey,))
    objectcache = None
    with globalcachelock:
        if cachekey in objectcache_dict:
            objectcache = objectcache_dict[cachekey]
            del objectcache_dict[cachekey]
    # Defer operations that acquire the cache local lock until 
    # the global lock is released
    if objectcache:
        objectcache.close()
    return

# ===== @@@Unused functions follow; eventually, remove these? ====

def make_cache_unused_(cachekey):
    """
    This function creates an object cache.

    cachekey    is a hashable value that uniquely identifies the required cache
                (e.g. a string or URI).

    Returns the created cache object.
    """
    with globalcachelock:
        try:
            if cachekey in objectcache_dict:
                raise Cache_Error(cachekey, msg="Cache already exists")
            objectcache_dict[cachekey] = ObjectCache(cachekey)
            objectcache_tb[cachekey]   = traceback.extract_stack()
            objectcache = objectcache_dict[cachekey]    # Copy value while lock acquired
        except Exception as e:
            print("@@@@ Cache already exists", file=sys.stderr)
            print("".join(traceback.format_list(objectcache_tb[cachekey])), file=sys.stderr)
            print("@@@@", file=sys.stderr)
            raise
    return objectcache

def remove_all_caches_unused_():
    """
    This function removes all caches from the set of object caches
    """
    # log.debug("@@@@ remove_all_caches")
    objectcaches = []
    with globalcachelock:
        for cachekey in objectcache_dict.keys():
            # log.debug("@@@@ remove_all_caches %r"%(cachekey,))
            objectcaches.append(objectcache_dict[cachekey])
            del objectcache_dict[cachekey]
    # Defer operations that acquire the cache local lock until 
    # the global lock is released
    for objectcache in objectcaches:
        objectcache.close()
    return

def remove_matching_caches_unused_(match_fn):
    """
    This function removes all caches whose cache key is matched by the provided
    function from the set of object caches.

    match_fn    is a function that tests a supplied cache key, and returns
                True if the corresponding cache is to be removed.
    """
    objectcaches = []
    with globalcachelock:
        for cachekey in _find_matching_cache_keys(match_fn):
            objectcaches.append(objectcache_dict[cachekey])
            del objectcache_dict[cachekey]
    # Defer operations that acquire the cache local lock until 
    # the global lock is released
    for objectcache in objectcaches:
        objectcache.close()
    return

def _find_matching_cache_keys_unused_(match_fn):
    """
    A generator that returns all cache keys matching the supplied function.

    match_fn    is a function that tests a supplied cache key, and returns
                True if it mnatches some criterion.
    """
    for cachekey in objectcache_dict.keys():
        if match_fn(cachekey):
            yield cachekey
    return

#   ===================================================================
#
#   Object cache class
#
#   ===================================================================

class ObjectCache(object):
    """
    A class for caching objects of some type.

    The cache is identified by is cache key value that is used to distinguish 
    a particular object cache from all others (see also `getCache`)
    """

    def __init__(self, cachekey):
        # log.debug("ObjectCache.__init__: cachekey %r"%(cachekey,))
        self._cachekey  = cachekey
        self._cachelock = threading.Lock()  # Allocate a lock object for this cache
        self._cache     = {}                # Initial empty set of values
        self._opened    = traceback.extract_stack()
        self._closed    = None
        return

    def cache_key(self):
        """
        Return cache key (e.g. for use with 'remove_cache')
        """
        return self._cachekey

    def flush(self):
        """
        Remove all objects from cache.
        """
        # log.debug("ObjectCache.flush: cachekey %r"%(self._cachekey,))
        with self._cachelock:
            for key in self._cache.keys():
                del self._cache[key]
        return self

    def close(self):
        """
        Close down this cache object.  Once closed, it cannot be used again.
        """
        # log.debug("ObjectCache.close: cachekey %r"%(self._cachekey,))
        self.flush()
        self._cachelock = None              # Discard lock object
        self._closed    = traceback.extract_stack()
        return

    def set(self, key, value):
        """
        Save object value in cache (overwriting any existing value for the key).

        key     is a hashable value that uniquely identifies the required cache
                (e.g. a string or URI).
        value   is a (new) value that is to be associated with the key.
        """
        with self._cachelock:
            self._cache[key] = value
        return value

    def get(self, key, default=None):
        """
        Retrieve object value from cache, or return default value
        """
        if self._cachelock is None:
            msg = "Access after cache closed (%r, %s)"%(self._cachekey, key)
            log.error(msg)
            log.debug("---- closed at:")
            log.debug("".join(traceback.format_list(self._closed)))
            log.debug("----")
            raise Exception(msg)
        # print("@@@@ self._cachelock %r, self._cachekey %r"%(self._cachelock, self._cachekey))
        with self._cachelock:
            value = self._cache.get(key, default)
        return value

    def pop(self, key, default=None):
        """
        Remove object value from cache, return that or default value
        """
        with self._cachelock:
            value = self._cache.pop(key, default)
        return value

    @contextlib.contextmanager
    def access(self, *keys):
        """
        A context manager for interlocked access to a cached value.

        The value bound by the context manager (for a 'with ... as' assignment) is a
        dictionary that has entries for each of the valuyes in the supplied key values
        for which there is a previously cached value.

        On exit from the context manager, if the value under any of the given keys has 
        been changed, or if any new entries have been added, they are used to update the 
        cached values before the interlock is released.

        Use like this:

            with cacheobject.access("key1", "key2", ...) as value:
                # value is dict of cached values for given keys
                # interlocked processing code here
                # updates to value are written back to cache on leavinbg context

        See: https://docs.python.org/2/library/contextlib.html
        """
        with self._cachelock:
            value_dict = {}
            for key in keys:
                if key in self._cache:
                    value_dict[key] = self._cache[key]
            yield value_dict
            for key in value_dict:
                self._cache[key] = value_dict[key]
            # If needed: this logic removes keys deleted by yield code...
            # for key in keys:
            #     if key not in value_dict:
            #         del self._cache[key]
        return

    # ===== @@@ UNUSED - remove this in due course =====

    def find_unused_(self, key, load_fn, seed_value):
        """
        Returns cached value for key, or calls the supplied function to obtain a 
        value, and caches and returns that value.

        If a previously-cached value is present, the value returned is:

            (False, old-value)

        If a previously-cached value is not present, the function is called with 
        the supplied "seed_value" as a parameter, the resuling value is cached under
        the supplied key, and the return value is:

            (True, new-value)
        """
        # log.debug("ObjectCache.find: cachekey %r, key %r"%(self._cachekey, key))
        with self._cachelock:
            if key in self._cache:
                old_value = self._cache[key]
                result = (False, old_value)
            else:
                new_value = load_fn(seed_value)
                self._cache[key] = new_value
                result = (True, new_value)
        return result

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for collection vocabulary namespace cache class.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from annalist.models.site                   import Site
from annalist.models.collection             import Collection
from annalist.models.recordvocab            import RecordVocab
from annalist.models.collectionvocabcache   import CollectionVocabCache

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site, init_annalist_test_coll, resetSitedata
    )
from .entity_testentitydata import (
    entitydata_create_values
    )

#   -----------------------------------------------------------------------------
#
#   CollectionVocabCache tests
#
#   -----------------------------------------------------------------------------

class CollectionVocabCacheTest(AnnalistTestCase):
    """
    Tests CollectionVocabCache class.
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite     = Site(TestBaseUri, TestBaseDir)
        self.testcoll     = Collection(self.testsite, "testcoll")
        self.testcoll1_a  = Collection(self.testsite, "testcoll1")
        self.testcoll1_b  = Collection(self.testsite, "testcoll1")
        self.testcoll2_a  = Collection(self.testsite, "testcoll2")
        self.testcoll2_b  = Collection(self.testsite, "testcoll2")
        self.vocabcache   = CollectionVocabCache()
        self.vocab1       = RecordVocab(self.testcoll1_a, "vocab1")
        self.vocab1.set_values(
            entitydata_create_values("vocab1", type_id="_vocab", entity_uri="test:vocab1#")
            )
        self.vocab2       = RecordVocab(self.testcoll1_a, "vocab2")
        self.vocab2.set_values(
            entitydata_create_values("vocab2", type_id="_vocab", entity_uri="test:vocab2#")
            )
        return

    def create_vocab_record(self, coll, vocab_entity):
        return RecordVocab.create(coll, vocab_entity.get_id(), vocab_entity.get_values())

    def create_test_vocab_entities(self):
        # Create entity records in collection data
        self.create_vocab_record(self.testcoll1_a, self.vocab1)
        self.create_vocab_record(self.testcoll1_a, self.vocab2)
        # Add entities to cache
        self.vocabcache.set_vocab(self.testcoll1_a, self.vocab1)
        self.vocabcache.set_vocab(self.testcoll1_a, self.vocab2)
        # Check cache content
        self.expect_site_vocab_ids = (
            { "owl", "rdf", "annal", "xsd", "rdfs"
            })
        self.expect_coll_vocab_ids = (
            {"vocab1", "vocab2"}
            )
        self.expect_all_vocab_ids = set.union(self.expect_site_vocab_ids, self.expect_coll_vocab_ids)
        retrieve_vocab_ids = set(t.get_id() for t in self.vocabcache.get_all_vocabs(self.testcoll1_a))
        self.assertEqual(retrieve_vocab_ids, self.expect_coll_vocab_ids)
        return

    def tearDown(self):
        resetSitedata(scope="collections")
        return

    def test_empty_cache(self):
        self.assertEqual(set(self.vocabcache.get_all_vocabs(self.testcoll1_a)), set())
        self.assertEqual(set(self.vocabcache.get_all_vocabs(self.testcoll1_b)), set())
        self.assertEqual(set(self.vocabcache.get_all_vocabs(self.testcoll2_a)), set())
        self.assertEqual(set(self.vocabcache.get_all_vocabs(self.testcoll2_b)), set())
        return

    def test_singleton_cache(self):
        # Test that cached values are shared between different instantiations
        # of the same collection.
        foo = self.vocabcache.get_vocab(self.testcoll1_a, "vocab2")
        self.assertEqual(self.vocabcache.get_vocab(self.testcoll1_a, "vocab2"), None)
        self.assertEqual(self.vocabcache.get_vocab(self.testcoll1_b, "vocab2"), None)
        self.assertEqual([t.get_uri() for t in self.vocabcache.get_all_vocabs(self.testcoll1_a)], [])
        self.create_vocab_record(self.testcoll1_a, self.vocab1)
        # NOTE: the 'set_vocab' call also causes cache initialization of all vocabs, 
        # including the created vocab record which is discovered on disk, and the 
        # subsequent set_vocab call returns 'False'.
        self.assertFalse(self.vocabcache.set_vocab(self.testcoll1_a, self.vocab1))
        self.assertFalse(self.vocabcache.set_vocab(self.testcoll1_b, self.vocab1))
        self.assertEqual(self.vocabcache.get_vocab(self.testcoll1_a, "vocab1").get_uri(), "test:vocab1#")
        self.assertEqual(self.vocabcache.get_vocab(self.testcoll1_b, "vocab1").get_uri(), "test:vocab1#")
        self.assertEqual(self.vocabcache.get_vocab(self.testcoll1_a, "vocab2"), None)
        self.assertEqual(self.vocabcache.get_vocab(self.testcoll1_b, "vocab2"), None)
        self.assertEqual(self.vocabcache.get_vocab_from_uri(self.testcoll1_a, "test:vocab1#").get_id(), "vocab1")
        self.assertEqual(self.vocabcache.get_vocab_from_uri(self.testcoll1_b, "test:vocab1#").get_id(), "vocab1")
        self.assertEqual(self.vocabcache.get_vocab_from_uri(self.testcoll1_a, "test:vocab2#"), None)
        self.assertEqual(self.vocabcache.get_vocab_from_uri(self.testcoll1_b, "test:vocab3#"), None)
        self.assertEqual([t.get_uri() for t in self.vocabcache.get_all_vocabs(self.testcoll1_a)], ["test:vocab1#"])
        self.assertEqual([t.get_uri() for t in self.vocabcache.get_all_vocabs(self.testcoll1_b)], ["test:vocab1#"])
        self.assertEqual([t.get_uri() for t in self.vocabcache.get_all_vocabs(self.testcoll2_a)], [])
        self.assertEqual([t.get_uri() for t in self.vocabcache.get_all_vocabs(self.testcoll2_b)], [])
        return

    def test_empty_cache_2(self):
        self.vocabcache.set_vocab(self.testcoll1_a, self.vocab1)
        self.assertIsNotNone(self.vocabcache.remove_vocab(self.testcoll1_a, "vocab1"))
        self.assertIsNone(self.vocabcache.remove_vocab(self.testcoll1_b, "vocab1"))
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll1_a)), [])
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll1_b)), [])
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll2_a)), [])
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll2_b)), [])
        return

    def test_flush_cache(self):
        self.vocabcache.set_vocab(self.testcoll1_a, self.vocab1)
        self.assertTrue(self.vocabcache.flush_cache(self.testcoll1_a))
        self.assertIsNone(self.vocabcache.remove_vocab(self.testcoll1_b, "vocab1"))
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll1_a)), [])
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll1_b)), [])
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll2_a)), [])
        self.assertEqual(list(self.vocabcache.get_all_vocabs(self.testcoll2_b)), [])
        self.assertTrue(self.vocabcache.flush_cache(self.testcoll1_b))
        self.assertFalse(self.vocabcache.flush_cache(self.testcoll1_b))
        return

    def test_get_vocabs_by_uri(self):
        self.create_test_vocab_entities()
        self.assertEqual(
            self.vocabcache.get_vocab_from_uri(self.testcoll1_b, "test:vocab1#").get_id(),
            "vocab1"
            )
        self.assertEqual(
            self.vocabcache.get_vocab_from_uri(self.testcoll1_b, "test:vocab2#").get_id(),
            "vocab2"
            )
        return

    def test_get_all_vocabs_scope_coll(self):
        self.create_test_vocab_entities()
        vocab_ids = set(t.get_id() for t in self.vocabcache.get_all_vocabs(self.testcoll1_a))
        self.assertEqual(vocab_ids, self.expect_coll_vocab_ids)
        vocab_ids = set(t.get_id() for t in self.vocabcache.get_all_vocabs(self.testcoll1_b))
        self.assertEqual(vocab_ids, self.expect_coll_vocab_ids)
        return

    def test_get_all_vocabs_scope_all(self):
        self.create_test_vocab_entities()
        vocab_ids = set(t.get_id() for t in self.vocabcache.get_all_vocabs(self.testcoll1_a, altscope="all"))
        self.assertEqual(vocab_ids, self.expect_all_vocab_ids)
        return

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for EntityData list view with additional inherited bibliography data
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.db                      import models
from django.http                    import QueryDict
# from django.contrib.auth.models     import User
from django.test                    import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client             import Client

from utils.SuppressLoggingContext   import SuppressLogging

from annalist                       import layout
from annalist.identifiers           import RDF, RDFS, ANNAL

from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitydata     import EntityData

from annalist.views.uri_builder             import uri_params, uri_with_params, continuation_params_url
from annalist.views.entitylist              import EntityGenericListView
from annalist.views.form_utils.fieldchoice  import FieldChoice

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    test_layout,
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site,
    init_annalist_test_coll,
    install_annalist_named_coll,
    create_test_coll_inheriting,
    init_annalist_named_test_coll,
    resetSitedata
    )
from .entity_testutils import (
    site_dir, collection_dir,
    site_view_url,
    collection_view_url, collection_edit_url,
    continuation_url_param,
    confirm_delete_params,
    collection_create_values,
    site_title,
    create_test_user, create_user_permissions,
    context_view_field,
    context_list_entities,
    context_list_head_fields, context_list_item_fields,
    context_list_item_field, context_list_item_field_value,
    check_context_field, check_context_field_value, check_context_list_field_value,
    check_field_list_context_fields,
    )
from .entity_testtypedata import (
    recordtype_dir, 
    recordtype_url,
    recordtype_create_values, 
    )
from .entity_testentitydata import (
    recorddata_dir,  entitydata_dir,
    entity_url, entitydata_edit_url, entitydata_delete_confirm_url,
    entitydata_list_type_url, entitydata_list_all_url,
    entitydata_value_keys, entitydata_create_values, entitydata_values, 
    default_view_form_data, entitydata_delete_confirm_form_data,
    entitylist_form_data
    )
from .entity_testsitedata import (
    make_field_choices, no_selection,
    get_site_bib_types, get_site_bib_types_sorted, get_site_bib_types_linked,
    get_site_bib_lists, get_site_bib_lists_sorted, get_site_bib_lists_linked,
    get_site_schema_types, get_site_schema_types_sorted, get_site_schema_types_linked,
    get_site_schema_lists, get_site_schema_lists_sorted, get_site_schema_lists_linked,
    )
from .entity_testlistdata import recordlist_url

#   -----------------------------------------------------------------------------
#
#   EntityDefaultListView tests
#
#   -----------------------------------------------------------------------------

class EntityInheritListViewTest(AnnalistTestCase):
    """
    Tests for get/copy/edit entity inherited from parent collection, 
    including tests of access control restrictions.
    """

    def setUp(self):
        self.testsite  = init_annalist_test_site()
        self.testcoll  = init_annalist_test_coll()
        self.testsubcoll = create_test_coll_inheriting(
            base_coll_id="testcoll", coll_id="testsubcoll", type_id="testtype"
            )
        create_test_user(self.testcoll,    "testuser",    "testpassword")
        create_test_user(self.testsubcoll, "testsubuser", "testpassword")
        # Allow user "testuser" access in collectrion "testsubcoll"
        user_permissions = ["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG"]
        user_id          = "testuser"
        user_perms = self.testsubcoll.create_user_permissions(
            user_id, "mailto:%s@%s"%(user_id, TestHost),
            "Test Subuser",
            "User %s: permissions for %s in collection %s"%
              ( user_id, "Test Subuser", self.testsubcoll.get_id() ),
            user_permissions)
        # Block user "testsubuser" access in collection "testcoll"
        user_permissions = []
        user_id          = "testsubuser"
        user_perms = self.testcoll.create_user_permissions(
            user_id, "mailto:%s@%s"%(user_id, TestHost),
            "Test Subuser",
            "User %s: permissions for %s in collection %s"%
              ( user_id, "Test Subuser", self.testcoll.get_id() ),
            user_permissions)
        # Block default user access in collection "testcoll"
        user_permissions = []
        user_id          = "_default_user_perms"
        user_uri         = "annal:User/_default_user_perms"
        user_perms = self.testcoll.create_user_permissions(
            user_id, user_uri,
            "Test Subuser",
            "User %s: permissions for %s in collection %s"%
              ( user_id, "Test Subuser", self.testcoll.get_id() ),
            user_permissions)
        # Create inherited entity "testcoll/testtype/entity2"
        self.testdata    = RecordTypeData.load(self.testcoll, "testtype")
        self.testentity2 = self._create_entity_data("entity2")
        self.testsubdata = RecordTypeData.load(self.testsubcoll, "testtype")
        # loggedin = self.client.login(username="testuser", password="testpassword")
        # self.assertTrue(loggedin)
        self.continuation_url = entitydata_list_type_url(
            coll_id="testsubcoll", type_id="testtype"
            )
        return

    def tearDown(self):
        resetSitedata(scope="collections")
        return

    @classmethod
    def setUpClass(cls):
        super(EntityInheritListViewTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(EntityInheritListViewTest, cls).tearDownClass()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _create_entity_data(self, entity_id, update="Entity"):
        "Helper function creates entity data in 'testcoll/testtype' with supplied id"
        e = EntityData.create(self.testdata, entity_id, 
            entitydata_create_values(entity_id, update=update)
            )
        return e    

    #   -----------------------------------------------------------------------------
    #   Form rendering tests
    #   -----------------------------------------------------------------------------

    #   -----------------------------------------------------------------------------
    #   Form submission tests
    #   -----------------------------------------------------------------------------

    #   See: test_upload_file.py

    #   -----------------------------------------------------------------------------
    #   Access control tests
    #   -----------------------------------------------------------------------------

    #   ----- Confirm view/edit/copy operations with access allowed -----

    # GET inherited entity from collection with access allowed

    def test_get_view_inherited_entity(self):
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        u = entitydata_edit_url(
            "edit", "testsubcoll", 
            "testtype", entity_id="entity2", 
            view_id="Default_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return

    # POST edit inherited entity from collection with access allowed

    def test_post_edit_inherited_entity(self):
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        f = default_view_form_data(action="edit", 
            entity_id="entity2", type_id="testtype", coll_id="testsubcoll",
            orig_coll="testcoll"
            )
        u = entitydata_edit_url(
            "edit", "testsubcoll", "testtype", entity_id="entity2", 
            view_id="Default_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertIn(self.continuation_url, r['location'])
        # Check that new data exists
        self.assertTrue(EntityData.exists(self.testsubdata, "entity2"))
        return

    # POST copy inherited entity from collection with access allowed

    def test_post_copy_inherited_entity(self):
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        f = default_view_form_data(action="copy", 
            entity_id="entity2", type_id="testtype", coll_id="testsubcoll",
            orig_coll="testcoll"
            )
        u = entitydata_edit_url(action="copy", 
            coll_id="testsubcoll", type_id="testtype", entity_id="entity2", 
            view_id="Default_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertIn(self.continuation_url, r['location'])
        # Check that new data exists
        self.assertTrue(EntityData.exists(self.testsubdata, "entity2"))
        return

    #   ----- Confirm view/edit/copy operations with access disallowed -----

    # GET inherited entity from collection with no access

    def test_post_edit_inherited_entity_no_access(self):
        loggedin = self.client.login(username="testsubuser", password="testpassword")
        # self.assertTrue(loggedin)
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        f = default_view_form_data(action="edit", 
            entity_id="entity2", type_id="testtype", coll_id="testsubcoll",
            orig_coll="testcoll"
            )
        u = entitydata_edit_url(
            "edit", "testsubcoll", "testtype", entity_id="entity2", 
            view_id="Default_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   403)
        self.assertEqual(r.reason_phrase, "Forbidden")
        # Check that no new data exists
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        return

    # POST edit inherited entity from collection with no access

    def test_post_edit_inherited_entity_no_access(self):
        loggedin = self.client.login(username="testsubuser", password="testpassword")
        self.assertTrue(loggedin)
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        f = default_view_form_data(action="edit", 
            entity_id="entity2", type_id="testtype", coll_id="testsubcoll",
            orig_coll="testcoll"
            )
        u = entitydata_edit_url(
            "edit", "testsubcoll", "testtype", entity_id="entity2", 
            view_id="Default_view"
            )
        r = self.client.post(u, f)
        # print "@@@@ r.content %r, r['location'] %r"%(r.content, r['location'])
        self.assertEqual(r.status_code,   403)
        self.assertEqual(r.reason_phrase, "Forbidden")
        # Check that no new data exists
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        return

    # POST copy inherited entity from collection with no access
    def test_post_copy_inherited_entity_no_access(self):
        loggedin = self.client.login(username="testsubuser", password="testpassword")
        self.assertTrue(loggedin)
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        f = default_view_form_data(action="copy", 
            entity_id="entity2", type_id="testtype", coll_id="testsubcoll",
            orig_coll="testcoll"
            )
        u = entitydata_edit_url(action="copy", 
            coll_id="testsubcoll", type_id="testtype", entity_id="entity2", 
            view_id="Default_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   403)
        self.assertEqual(r.reason_phrase, "Forbidden")
        # Check that no new data exists
        self.assertFalse(EntityData.exists(self.testsubdata, "entity2"))
        return

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Collection resource access (JSON-LD context and maybe more)
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import logging
log = logging.getLogger(__name__)

from django.http                        import HttpResponse

from annalist                           import message
from annalist                           import layout

import annalist.models.entitytypeinfo as entitytypeinfo
from annalist.models.entityresourceaccess import (
    collection_fixed_json_resources,
    find_entity_resource,
    get_resource_file
    )

from annalist.views.displayinfo         import DisplayInfo
from annalist.views.generic             import AnnalistGenericView

class CollectionResourceAccess(AnnalistGenericView):
    """
    View class for collection resource access

    This view class returns a collection data resource.
    """

    def __init__(self):
        super(CollectionResourceAccess, self).__init__()
        return

    # GET

    def get(self, request, 
            coll_id=None, resource_ref=None, view_id=None):
        """
        Access specified entity resource
        """
        log.debug(
            "CollectionResourceAccess.get: coll_id %s, resource_ref %s"%
            (coll_id, resource_ref)
            )
        viewinfo = self.view_setup(
            coll_id, request.GET.dict()
            )
        if viewinfo.http_response:
            return viewinfo.http_response

        coll = viewinfo.collection 
        if coll is None:
            return self.error(
                dict(self.error404values(),
                    message=message.COLLECTION_NOT_EXISTS%{'id': coll_id}
                    )
                )
        # Locate resource
        resource_info = find_entity_resource(
            viewinfo.site, resource_ref, fixed_resources=collection_fixed_json_resources
            )
        log.debug("CollectionResourceAccess.get: resource_info %r"%(resource_info,))
        if resource_info is None:
            return self.error(
                dict(self.error404values(),
                    message=message.COLL_RESOURCE_NOT_DEFINED%
                        { 'id':  coll_id
                        , 'ref': resource_ref
                        }
                    )
                )
        coll_baseurl = viewinfo.reqhost + self.get_collection_base_url(coll_id)
        resource_file, resource_type = get_resource_file(
            coll, resource_info, coll_baseurl
            )
        if resource_file is None:
            return self.error(
                dict(self.error404values(),
                    message=message.COLL_RESOURCE_NOT_EXIST%
                        { 'id':  coll_id
                        , 'ref': resource_ref
                        }
                    )
                )
        # Return resource
        try:
            response = self.resource_response(resource_file, resource_info["resource_type"])
        except Exception as e:
            log.exception(str(e))
            response = self.error(
                dict(self.error500values(),
                    message=str(e)+" - see server log for details"
                    )
                )
        finally:
            resource_file.close()
        return response

    def view_setup(self, coll_id, request_dict):
        """
        Assemble display information for view request handler
        """
        action                        = "view"
        self.default_continuation_url = None
        viewinfo = DisplayInfo(self, action, request_dict, self.default_continuation_url)
        viewinfo.get_site_info(self.get_request_host())
        viewinfo.get_coll_info(coll_id)
        viewinfo.get_request_type_info(entitytypeinfo.COLL_ID)
        viewinfo.check_authorization(action)
        return viewinfo

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Entity list data views
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2015, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import json

import logging
log = logging.getLogger(__name__)

from django.conf                        import settings
from django.http                        import HttpResponse
from django.http                        import HttpResponseRedirect

from annalist                           import message
from annalist                           import layout
from annalist.identifiers               import RDFS, ANNAL
from annalist.util                      import make_type_entity_id

from annalist.models.entityfinder       import EntityFinder
from annalist.models.entityresourceaccess import (
    find_list_resource,
    json_resource_file
    )

from annalist.views.entitylist          import EntityGenericListView

#   -------------------------------------------------------------------------------------------
#
#   List entities view - form rendering and POST response handling
#
#   -------------------------------------------------------------------------------------------

class EntityListDataView(EntityGenericListView):
    """
    View class for generic entity list returned as JSON-LD
    """

    def __init__(self):
        super(EntityListDataView, self).__init__()
        return

    # GET

    def get(self, request, coll_id=None, type_id=None, list_id=None, list_ref=None):
        """
        Return a list of entities as a JSON-LD object

        NOTE: The current implementation returns a full copy of each of the 
        selected entities.
        """
        scope      = request.GET.get('scope',  None)
        search_for = request.GET.get('search', "")
        log.info(
            "views.entitylistdata.get: coll_id %s, type_id %s, list_id %s, list_ref %s, scope %s, search %s"%
            (coll_id, type_id, list_id, list_ref, scope, search_for)
            )
        listinfo    = self.list_setup(coll_id, type_id, list_id, request.GET.dict())
        if listinfo.http_response:
            return listinfo.http_response
        # print "@@@@ listinfo.type_id %s, type_id %s"%(listinfo.type_id, type_id)
        # log.debug("@@ listinfo.list_id %s, coll base_url %s"%(listinfo.list_id, base_url))
        # Prepare list data for rendering
        try:
            jsondata = self.assemble_list_data(listinfo, scope, search_for)
        except Exception as e:
            log.exception(str(e))
            return self.error(
                dict(self.error500values(),
                    message=str(e)+" - see server log for details"
                    )
                )
        entity_list_info = find_list_resource(type_id, list_id, list_ref)
        if entity_list_info is None:
            return self.error(
                dict(self.error404values(),
                    message=message.LIST_NOT_DEFINED%
                        { 'list_id':  list_id
                        , 'type_id':  type_id
                        , 'list_ref': list_ref
                        }
                    )
                )

        coll_baseurl = listinfo.reqhost + self.get_collection_base_url(coll_id)
        list_baseurl = listinfo.reqhost + self.get_list_base_url(coll_id, type_id, list_id)
        if "resource_access" in entity_list_info:
            # Use indicated resource access renderer
            list_file_access = entity_list_info["resource_access"]
        else:
            list_file_access = json_resource_file
        list_file = list_file_access(list_baseurl, jsondata, entity_list_info)
        if list_file is None:
            return self.error(
                dict(self.error404values(),
                    message=message.LIST_NOT_ACCESSED%
                        { 'list_id':  list_id
                        , 'type_id':  type_id
                        , 'list_ref': list_ref
                        }
                    )
                )

        # Construct and return list response
        try:
            return_type = entity_list_info["resource_type"]
            # URL parameter ?type=mime/type overrides specified content type
            #
            # @@TODO: this is to allow links to return different content-types:
            #         is there a cleaner way?
            if "type" in listinfo.request_dict:
                return_type = listinfo.request_dict["type"]
            links=[
                { "rel": "canonical"
                , "ref": list_baseurl
                }]
            response = self.resource_response(list_file, return_type, links=links)
        except Exception as e:
            log.exception(str(e))
            response = self.error(
                dict(self.error500values(),
                    message=str(e)+" - see server log for details"
                    )
                )
        finally:
            list_file.close()

        return response

# End.



# !/usr/bin/env python
#
# am_main.py - command line tool to perform Annalist installation management
#

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2013-2014, Graham Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import os.path
import re
import argparse
import logging
import errno

log = logging.getLogger(__name__)

# if __name__ == "__main__":
dirhere = os.path.dirname(os.path.realpath(__file__))
annroot = os.path.dirname(os.path.join(dirhere))
sys.path.insert(0, annroot)
# sys.path.insert(0, dirhere)

import annalist

from .                      import am_errors
from .am_runtests           import am_runtests
from .am_initialize         import am_initialize, am_collectstatic
from .am_createsite         import am_createsite, am_updatesite
from .am_runserver          import (
    am_runserver, am_stopserver, am_pidserver, am_rundevserver, 
    am_serverlog, am_accesslog, am_errorlog, am_sitedirectory, 
    am_settingsmodule, am_settingsfile, am_settingsdir, 
    am_version
    )
from .am_createuser         import (
    am_createlocaluser, am_createadminuser, am_defaultadminuser, am_updateadminuser, 
    am_setuserpermissions, am_setdefaultpermissions, am_setpublicpermissions,
    am_deleteuser
    )
from .am_managecollections  import (
    am_installcollection, am_copycollection,
    am_migrationreport, am_migratecollection, am_migrateallcollections
    )
from .am_help               import am_help, command_summary_help

VERSION = annalist.__version__

def progname(args):
    return os.path.basename(args[0])

def parseCommandArgs(argv):
    """
    Parse command line arguments

    argv            argument list from command line

    Returns a pair consisting of options specified as returned by
    OptionParser, and any remaining unparsed arguments.
    """
    # create a parser for the command line options
    parser = argparse.ArgumentParser(
                description="Annalist site management utility",
                formatter_class=argparse.RawDescriptionHelpFormatter,
                epilog=command_summary_help
                )
    parser.add_argument('--version', action='version', version='%(prog)s '+VERSION)
    parser.add_argument("-c", "--configuration",
                        action='store',
                        dest="configuration", metavar="CONFIG",
                        default="personal",
                        #choices=['personal', 'shared', 'devel', 'runtests'],
                        help="Select site configuration by name (e.g. personal, shared, devel, runtests.")
    parser.add_argument("-p", "--personal",
                        action='store_true',
                        dest="config_p", # metavar="PERSONAL",
                        help="Select personal site configuration.")
    parser.add_argument("-d", "--development",
                        action='store_true',
                        dest="config_d", # metavar="DEVELOPMENT",
                        help="Select development site configuration.")
    parser.add_argument("-s", "--shared",
                        action='store_true',
                        dest="config_s", # metavar="SHARED",
                        help="Select shared site configuration.")
    parser.add_argument("-f", "--force",
                        action='store_true',
                        dest="force",
                        help="Force overwrite of existing site data or collection.")
    parser.add_argument("--debug",
                        action="store_true", 
                        dest="debug", 
                        default=False,
                        help="Run with full debug output enabled.  "+
                             "Also creates log file 'annalist-manager.log' in the working directory"
                        )
    parser.add_argument("command", metavar="COMMAND",
                        nargs=None,
                        help="sub-command, one of the options listed below."
                        )
    parser.add_argument("args", metavar="ARGS",
                        nargs="*",
                        help="Additional arguments, depending on the command used."
                        )
    # parse command line now
    options = parser.parse_args(argv)
    if options:
        if options.config_d: options.configuration = "devel"
        if options.config_p: options.configuration = "personal"
        if options.config_s: options.configuration = "shared"
        if options and options.command:
            return options
    print("No valid usage option given.", file=sys.stderr)
    parser.print_usage()
    return None

def run(userhome, userconfig, options, progname):
    """
    Command dispatcher.
    """
    if options.command.startswith("runt"):                  # runtests
        return am_runtests(annroot, options)
    if options.command.startswith("init"):                  # initialize (intsllaation, django database)
        return am_initialize(annroot, userhome, userconfig, options)
    if options.command.startswith("collect"):               # collect static data
        return am_collectstatic(annroot, userhome, userconfig, options)
    if options.command.startswith("createl"):               # createlocaluser
        return am_createlocaluser(annroot, userhome, options)
    if options.command.startswith("createa"):               # createadminuser
        return am_createadminuser(annroot, userhome, options)
    if options.command.startswith("defaulta"):              # defaultadminuser
        return am_defaultadminuser(annroot, userhome, options)
    if options.command.startswith("updatea"):               # updateadminuser
        return am_updateadminuser(annroot, userhome, options)
    if options.command.startswith("setuse"):                # setuserpermissions
        return am_setuserpermissions(annroot, userhome, options)
    if options.command.startswith("setdef"):                # setdefaultpermissions
        return am_setdefaultpermissions(annroot, userhome, options)
    if options.command.startswith("setpub"):                # setpublicpermissions
        return am_setpublicpermissions(annroot, userhome, options)
    if options.command.startswith("deleteu"):               # deleteuser
        return am_deleteuser(annroot, userhome, options)
    if options.command.startswith("creates"):               # createsitedata
        return am_createsite(annroot, userhome, options)
    if options.command.startswith("updates"):               # updatesitedata
        return am_updatesite(annroot, userhome, options)
    if options.command.startswith("installc"):              # installcollection
        return am_installcollection(annroot, userhome, options)
    if options.command.startswith("copyc"):                 # copycollection
        return am_copycollection(annroot, userhome, options)
    if options.command.startswith("migrationr"):            # migrationreport
        return am_migrationreport(annroot, userhome, options)
    if options.command.startswith("migratec"):              # migratecollection
        return am_migratecollection(annroot, userhome, options)
    if options.command.startswith("migratea"):              # migrateallcollections
        return am_migrateallcollections(annroot, userhome, options)
    if options.command.startswith("runs"):                  # runserver
        return am_runserver(annroot, userhome, options)
    if options.command.startswith("stop"):                  # stopserver
        return am_stopserver(annroot, userhome, options)
    if options.command.startswith("pid"):                   # pidserver
        return am_pidserver(annroot, userhome, options)
    if options.command.startswith("rund"):                  # rundevserver
        return am_rundevserver(annroot, userhome, options)
    if options.command.startswith("serv"):                  # serverlog
        return am_serverlog(annroot, userhome, options)
    if options.command.startswith("acc"):                   # accesslog
        return am_accesslog(annroot, userhome, options)
    if options.command.startswith("err"):                   # errorlog
        return am_errorlog(annroot, userhome, options)
    if options.command.startswith("site"):                  # sitedir
        return am_sitedirectory(annroot, userhome, options)
    if options.command.startswith("settingsm"):             # settingsmodule
        return am_settingsmodule(annroot, userhome, options)
    if options.command.startswith("settingsf"):             # settingsfile
        return am_settingsfile(annroot, userhome, options)
    if options.command.startswith("settingsd"):             # settingsdir
        return am_settingsdir(annroot, userhome, options)
    if options.command.startswith("ver"):                   # version
        return am_version(annroot, userhome, options)
    if options.command.startswith("help"):
        return am_help(options, progname)
    print("Un-recognised sub-command: %s"%(options.command), file=sys.stderr)
    print("Use '%s --help' to see usage summary"%(progname), file=sys.stderr)        
    return am_errors.AM_BADCMD

def runCommand(userhome, userconfig, argv):
    """
    Run program with supplied configuration base directory, 
    configuration directory and command arguments.

    This is called by main function (below), and also by test suite routines.

    Returns exit status.
    """
    options = parseCommandArgs(argv[1:])
    # if options and options.debug:
    #     # This doesn't work as expected - some kind of interference with Django log settings?
    #     logging.basicConfig(level=logging.DEBUG, filename="annalist-manager.log", filemode="w")
    # else:
    #     logging.basicConfig(level=logging.INFO)
    # log.debug("runCommand: configbase %s, filebase %s, argv %s"%(configbase, filebase, repr(argv)))
    # log.debug("Options: %s"%(repr(options)))
    # else:
    #     logging.basicConfig()
    if options:
        progname = os.path.basename(argv[0])
        status   = run(userhome, userconfig, options, progname)
    else:
        status = am_errors.AM_BADCMD
    return status

def runMain():
    """
    Main program transfer function for setup.py console script
    """
    userhome   = os.path.expanduser("~")
    userconfig = os.path.join(userhome, ".annalist")
    return runCommand(userhome, userconfig, sys.argv)

if __name__ == "__main__":
    """
    Program invoked from the command line.
    """
    p = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.insert(0, p)
    status = runMain()
    sys.exit(status)

# End.

